"""This is the preprocessing submodule.

This module contains the functionality necessary to preprocess the raw data.
This includes calibration, deriving the various metrics, and the sleep
and physical activity analysis tools.
"""
